# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022-2024 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
from pathlib import Path
from typing import Any, Optional

from flask import Flask, jsonify, render_template
from flask.wrappers import Response

from .__version__ import __version__
from .gtfs import Gtfs, Row
from .tables import agency, calendar, calendar_dates, frequencies, routes, stops, times, trips
from .util import time_to_int, to_js_literal


class Application:
    def __init__(self, gtfs: Gtfs) -> None:
        self.gtfs = gtfs
        self.flask = Flask(__name__)
        self._init_app()

    def _init_app(self) -> None:
        self._init_template_functions()
        self._init_html_routes()
        self._init_api_routes()

    def _init_template_functions(self) -> None:
        # Apply template filters
        self.flask.add_template_global(agency.header_class, "agency_header_class")
        self.flask.add_template_global(routes.header_class, "routes_header_class")
        self.flask.add_template_global(routes.format_cell, "routes_format_cell")
        self.flask.add_template_global(stops.header_class, "stops_header_class")
        self.flask.add_template_global(stops.format_cell, "stops_format_cell")
        self.flask.add_template_global(trips.header_class, "trips_header_class")
        self.flask.add_template_global(trips.format_cell, "trips_format_cell")
        self.flask.add_template_global(times.header_class, "times_header_class")
        self.flask.add_template_global(times.format_cell, "times_format_cell")
        self.flask.add_template_global(frequencies.header_class, "frequencies_header_class")
        self.flask.add_template_global(frequencies.format_cell, "frequencies_format_cell")
        self.flask.add_template_global(calendar.header_class, "calendar_header_class")
        self.flask.add_template_global(calendar.format_cell, "calendar_format_cell")
        self.flask.add_template_global(calendar_dates.header_class, "calendar_dates_header_class")
        self.flask.add_template_global(calendar_dates.format_cell, "calendar_dates_format_cell")

        # Helper functions
        self.flask.add_template_global(to_js_literal, "to_js_literal")
        self.flask.add_template_global(
            lambda trip_id: self.gtfs.stop_times[trip_id][0]["departure_time"],  # type: ignore
            "trip_first_time",
        )
        self.flask.add_template_global(
            lambda trip_id: self.gtfs.stop_times[trip_id][-1]["departure_time"],  # type: ignore
            "trip_last_time",
        )

    def _init_html_routes(self) -> None:
        self.flask.add_url_rule("/", view_func=self.route_agency)
        self.flask.add_url_rule("/agency", view_func=self.route_agency)
        self.flask.add_url_rule("/agency/<path:agency_id>", view_func=self.route_routes)
        self.flask.add_url_rule("/routes", view_func=self.route_routes)
        self.flask.add_url_rule("/stops", view_func=self.route_stops)
        self.flask.add_url_rule("/route/<path:route_id>", view_func=self.route_trips)
        self.flask.add_url_rule("/block/<path:block_id>", view_func=self.route_trips)
        self.flask.add_url_rule("/stop/<path:stop_id>", view_func=self.route_stop)
        self.flask.add_url_rule("/trip/<path:trip_id>", view_func=self.route_trip)
        self.flask.add_url_rule("/calendars", view_func=self.route_calendars)
        self.flask.add_url_rule("/calendar/<path:service_id>", view_func=self.route_calendar)

    def _init_api_routes(self) -> None:
        self.flask.add_url_rule("/api/map/stops", view_func=self.route_api_map_stops)
        self.flask.add_url_rule("/api/map/stop/<path:stop_id>", view_func=self.route_api_map_stop)
        self.flask.add_url_rule("/api/map/trip/<path:trip_id>", view_func=self.route_api_map_trip)
        self.flask.add_url_rule(
            "/api/map/shape/<path:shape_id>",
            view_func=self.route_api_map_shape,
        )
        self.flask.add_url_rule(
            "/api/calendar/days/<path:service_id>",
            view_func=self.route_api_calendar_dates,
        )

    # HTML routes

    def route_agency(self) -> str:
        return render_template(
            "agency.html.jinja",
            missing=not self.gtfs.agency,
            header=self.gtfs.header_of("agency"),
            data=self.gtfs.agency.values(),
        )

    def route_routes(self, agency_id: Optional[str] = None) -> str:
        if agency_id:
            data = filter(lambda row: row["agency_id"] == agency_id, self.gtfs.routes.values())
        else:
            data = self.gtfs.routes.values()

        return render_template(
            "routes.html.jinja",
            missing=not self.gtfs.routes,
            header=self.gtfs.header_of("routes"),
            data=data,
        )

    def route_stops(self) -> str:
        return render_template(
            "stops.html.jinja",
            missing=not self.gtfs.stops,
            header=self.gtfs.header_of("stops"),
            data=self.gtfs.stops.values(),
        )

    def route_trips(self, route_id: Optional[str] = None, block_id: Optional[str] = None) -> str:
        if route_id:
            data = filter(lambda row: row["route_id"] == route_id, self.gtfs.trips.values())
        elif block_id:
            data = filter(lambda row: row["block_id"] == block_id, self.gtfs.trips.values())
        else:
            raise RuntimeError("Trips view must be filtered by either a block_id or route_id")

        return render_template(
            "trips.html.jinja",
            missing=not self.gtfs.trips,
            header=self.gtfs.header_of("trips"),
            data=data,
        )

    def route_stop(self, stop_id: str) -> str:
        # Special case for missing stops
        if stop_id not in self.gtfs.stops:
            return render_template(
                "stop.html.jinja",
                missing=True,
                stop={"stop_id": stop_id},
            )

        stop = self.gtfs.stops[stop_id]

        # Gather the stop_times of this stop,
        # and their trip_short_names and trip_headsigns
        times_by_service: dict[str, list[Row]] = {}
        trip_short_names: Optional[dict[str, str]] = (
            {} if "trip_short_name" in self.gtfs.header_of("trips") else None
        )
        trip_headsigns: Optional[dict[str, str]] = (
            {} if "trip_headsign" in self.gtfs.header_of("trips") else None
        )

        for time in self.gtfs.stop_times_by_stops.get(stop_id, []):
            if time["trip_id"] not in self.gtfs.trips:
                continue
            trip = self.gtfs.trips[time["trip_id"]]

            times_by_service.setdefault(trip["service_id"], []).append(time)

            if trip_short_names is not None:
                trip_short_names[trip["trip_id"]] = trip["trip_short_name"]
            if trip_headsigns is not None:
                trip_headsigns[trip["trip_id"]] = trip["trip_headsign"]

        # Ensure times_by_service are ordered by the departure time
        for lst in times_by_service.values():
            lst.sort(key=lambda t: time_to_int(t.get("departure_time", "")))

        # Calculate colspan
        colspan = len(self.gtfs.header_of("stop_times"))
        colspan += trip_short_names is not None
        colspan += trip_headsigns is not None

        # Render the template
        return render_template(
            "stop.html.jinja",
            missing=False,
            stop=stop,
            stops_header=self.gtfs.header_of("stops"),
            stops_in_group=self.gtfs.all_stops_in_group(stop_id),
            times_header=self.gtfs.header_of("stop_times"),
            times_by_service=times_by_service,
            trip_short_names=trip_short_names,
            trip_headsigns=trip_headsigns,
            times_colspan=colspan,
        )

    def route_trip(self, trip_id: str) -> str:
        # Short circuit for missing trips
        if trip_id not in self.gtfs.trips:
            return render_template(
                "trip.html.jinja",
                missing=True,
                trip={"trip_id": trip_id},
            )

        # Prepare data for rendering
        trip = self.gtfs.trips[trip_id]
        times = self.gtfs.stop_times.get(trip_id, [])

        stop_names = [
            self.gtfs.stops.get(i.get("stop_id", ""), {}).get("stop_name", "") for i in times
        ]

        return render_template(
            "trip.html.jinja",
            missing=False,
            trip=trip,
            trips_header=self.gtfs.header_of("trips"),
            times=times,
            times_header=self.gtfs.header_of("stop_times"),
            stop_names=stop_names,
            frequencies=self.gtfs.frequencies.get(trip_id),
            frequencies_header=self.gtfs.header_of("frequencies"),
        )

    def route_calendars(self) -> str:
        return render_template(
            "calendars.html.jinja",
            missing=(not self.gtfs.calendar) and (not self.gtfs.calendar_dates),
            header=self.gtfs.header_of("calendar"),
            data=self.gtfs.calendar.values(),
            implicit_calendars=[
                i for i in self.gtfs.calendar_dates if i not in self.gtfs.calendar
            ],
        )

    def route_calendar(self, service_id: str) -> str:
        missing = (
            service_id not in self.gtfs.calendar and service_id not in self.gtfs.calendar_dates
        )

        return render_template(
            "calendar.html.jinja",
            service_id=service_id,
            missing=missing,
            calendar_row=self.gtfs.calendar.get(service_id),
            calendar_header=self.gtfs.header_of("calendar"),
            calendar_dates_rows=self.gtfs.calendar_dates.get(service_id),
            calendar_dates_header=self.gtfs.header_of("calendar_dates"),
        )

    # JSON routes for map presentation

    def route_api_map_stops(self) -> Response:
        return jsonify(
            [
                {
                    "id": stop.get("stop_id"),
                    "code": stop.get("stop_code"),
                    "name": stop.get("stop_name"),
                    "lat": stop.get("stop_lat"),
                    "lon": stop.get("stop_lon"),
                }
                for stop in self.gtfs.stops.values()
            ]
        )

    def route_api_map_stop(self, stop_id: str) -> Response:
        return jsonify(
            [
                {
                    "idx": idx,
                    "id": stop.get("stop_id"),
                    "code": stop.get("stop_code"),
                    "name": stop.get("stop_name"),
                    "lat": stop.get("stop_lat"),
                    "lon": stop.get("stop_lon"),
                }
                for idx, stop in enumerate(self.gtfs.all_stops_in_group(stop_id))
            ]
        )

    def route_api_map_trip(self, trip_id: str) -> Response:
        stop_to_sequences: dict[str, list[str]] = {}
        for time in self.gtfs.stop_times.get(trip_id, []):
            stop_to_sequences.setdefault(time["stop_id"], []).append(time["stop_sequence"])

        stops: list[dict[str, Any]] = []
        for stop_id, sequences in stop_to_sequences.items():
            stop = self.gtfs.stops.get(stop_id, {})
            stops.append(
                {
                    "id": stop_id,
                    "lat": stop.get("stop_lat"),
                    "lon": stop.get("stop_lon"),
                    "name": stop.get("stop_name"),
                    "seq": sequences,
                }
            )

        return jsonify(stops)

    def route_api_map_shape(self, shape_id: str) -> Response:
        return jsonify(self.gtfs.shapes.get(shape_id, []))

    # JSON calendar data

    def route_api_calendar_dates(self, service_id: str) -> Response:
        return jsonify([i.isoformat() for i in self.gtfs.all_dates_of(service_id)])

    # Main entry point

    def run(self, debug: bool = False) -> None:
        return self.flask.run(load_dotenv=False, debug=debug, use_evalex=False)


def make_app() -> Flask:
    # Parse the arguments
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("file", type=Path, help="path to GTFS directory/zip")
    args = arg_parser.parse_args()

    # Load GTFS data
    gtfs = Gtfs.from_user_input(args.file)

    # Create the application
    app = Application(gtfs)

    # Return the created Flask instance
    return app.flask


def main() -> int:
    # Parse the arguments
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("file", type=Path, help="path to GTFS directory/zip")
    arg_parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="enable debug mode in Flask",
    )
    arg_parser.add_argument("-V", "--version", action="version", version=f"jvig {__version__}")
    args = arg_parser.parse_args()

    # Load GTFS data
    gtfs = Gtfs.from_user_input(args.file)

    # Create the application
    app = Application(gtfs)

    # Run it
    app.run(args.debug)
    return 0



__title__ = "jvig"
__description__ = "GTFS Viewer, created using Flask."
__version__ = "1.1.1"
__url__ = "https://github.com/MKuranowski/jvig"
__author__ = "Mikołaj Kuranowski"
__email__ = "mkuranowski+pypackages@gmail.com"
__copyright__ = "© Copyright 2020-2024 Mikołaj Kuranowski"
__license__ = "GNU General Public License v3.0 or later"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from markupsafe import escape

from .. import valid

VALID_FIELDS: set[str] = {
    "stop_id",
    "stop_code",
    "stop_name",
    "stop_desc",
    "stop_lat",
    "stop_lon",
    "zone_id",
    "stop_url",
    "location_type",
    "parent_station",
    "stop_timezone",
    "wheelchair_boarding",
    "platform_code",
}


def header_class(field: str) -> str:
    return "" if field in VALID_FIELDS else "value-unrecognized"


def format_cell(row: dict[str, str], field: str) -> str:
    value = row[field]

    if field == "stop_lat":
        if valid.latitude(value) is not None:
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "stop_lon":
        if valid.longitude(value) is not None:
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "location_type":
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (🚏)</td>"
        elif value == "1":
            return "<td>1 (🏢)</td>"
        elif value == "2":
            return "<td>2 (➡️🚪)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "wheelchair_boarding":
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (♿❓)</td>"
        elif value == "1":
            return "<td>1 (♿✔️)</td>"
        elif value == "2":
            return "<td>2 (♿❌)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    else:
        return f"<td>{escape(value)}</td>"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from markupsafe import escape

from .. import valid

WEEKDAYS: set[str] = {
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
}

VALID_FIELDS: set[str] = {"service_id", "start_date", "end_date"} | WEEKDAYS

EXTENDED_FIELDS: set[str] = {"service_desc"}


def header_class(field: str) -> str:
    if field in EXTENDED_FIELDS:
        return "value-extended"
    elif field in VALID_FIELDS:
        return ""
    return "value-unrecognized"


def format_cell(row: dict[str, str], field: str) -> str:
    value = row[field]

    if field in {"start_date", "end_date"}:
        if valid.date(value):
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field in WEEKDAYS:
        if value == "0":
            return "<td>0 (❌)</td>"
        elif value == "1":
            return "<td>1 (✔️)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    else:
        return f"<td>{escape(value)}</td>"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

VALID_FIELDS: set[str] = {
    "agency_id",
    "agency_name",
    "agency_url",
    "agency_timezone",
    "agency_lang",
    "agency_phone",
    "agency_fare_url",
    "agency_email",
}


def header_class(field: str) -> str:
    return "" if field in VALID_FIELDS else "value-unrecognized"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import quote_plus

from markupsafe import escape

VALID_FIELDS: set[str] = {
    "route_id",
    "service_id",
    "trip_id",
    "trip_headsign",
    "trip_short_name",
    "direction_id",
    "block_id",
    "shape_id",
    "wheelchair_accessible",
    "bikes_allowed",
}

EXTENDED_FIELDS: set[str] = {"exceptional"}


def header_class(field: str) -> str:
    if field in EXTENDED_FIELDS:
        return "value-extended"
    elif field in VALID_FIELDS:
        return ""
    return "value-unrecognized"


def format_cell(row: dict[str, str], field: str) -> str:
    value = row[field]

    if field == "route_id":
        return f'<td><a href="/route/{quote_plus(value)}">{escape(value)}</a></td>'

    elif field == "service_id":
        return f'<td><a href="/calendar/{quote_plus(value)}">{escape(value)}</a></td>'

    elif field in {"direction_id", "exceptional"}:
        if value in {"0", "1"}:
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "block_id":
        return f'<td><a href="/block/{quote_plus(value)}">{escape(value)}</a></td>'

    elif field == "wheelchair_accessible":
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (♿❓)</td>"
        elif value == "1":
            return "<td>1 (♿✔️)</td>"
        elif value == "2":
            return "<td>2 (♿❌)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "bikes_allowed":
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (🚲❓)</td>"
        elif value == "1":
            return "<td>1 (🚲✔️)</td>"
        elif value == "2":
            return "<td>2 (🚲❌)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    else:
        return f"<td>{escape(value)}</td>"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import re
from math import isnan
from typing import Optional

COLOR_PATTERN = re.compile(r"^[0-9A-Fa-f]{6}$")
DATE_PATTERN = re.compile(r"^[0-9]{8}$")
TIME_PATTERN = re.compile(r"^([0-9]+):([0-9]{2}):([0-9]{2})$")

DECIMAL_PATTERN = re.compile(r"^-?[0-9]+(?:\.[0-9]+)?$")
UINT_PATTERN = re.compile(r"^[0-9]+$")


def color(text: str) -> bool:
    return not text or COLOR_PATTERN.match(text) is not None


def latitude(text: str) -> Optional[float]:
    if not DECIMAL_PATTERN.match(text):
        return None

    value = float(text)

    if isnan(value) or value > 90.0 or value < -90.0:
        return None

    return value


def longitude(text: str) -> Optional[float]:
    if not DECIMAL_PATTERN.match(text):
        return None

    value = float(text)

    if isnan(value) or value > 180.0 or value < -180.0:
        return None

    return value


def time(text: str) -> bool:
    return TIME_PATTERN.match(text) is not None


def non_negative_float(text: str) -> bool:
    return not text or (text[:1] != "-" and DECIMAL_PATTERN.match(text) is not None)


def uint(text: str) -> bool:
    return UINT_PATTERN.match(text) is not None


def date(text: str) -> bool:
    if DATE_PATTERN.match(text) is None:
        return False
    try:
        datetime.date(
            int(text[0:4]),
            int(text[4:6]),
            int(text[6:8]),
        )
        return True
    except ValueError:
        return False



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022-2023 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import date
from io import StringIO
from pathlib import Path
from typing import ClassVar, Optional

from jvig.gtfs import Gtfs

FIXTURE_PATH = Path(__file__).with_name("fixtures")

# TODO: Test stop-station structures
# TODO: Test missing agency_id
# TODO: Test Gtfs.all_stops_in_group


class BaseWkdGtfsTest:
    def get_gtfs(self) -> Gtfs:
        raise NotImplementedError("Subclasses of BaseGtfsTest must implement get_gtfs()")

    def test_single_row(self) -> None:
        g = self.get_gtfs()
        assert len(g.agency) == 1
        assert "0" in g.agency

        a = g.agency["0"]
        assert a["agency_id"] == "0"
        assert a["agency_name"] == "Warszawska Kolej Dojazdowa"
        assert a["agency_url"] == "http://www.wkd.com.pl/"
        assert a["agency_lang"] == "pl"
        assert a["agency_timezone"] == "Europe/Warsaw"

    def test_many_rows(self) -> None:
        g = self.get_gtfs()
        assert len(g.calendar_dates) == 2
        assert "C" in g.calendar_dates
        assert "D" in g.calendar_dates

        ds = g.calendar_dates["C"]
        assert len(ds) == 9

        d = ds[2]
        assert d["service_id"] == "C"
        assert d["date"] == "20220417"
        assert d["exception_type"] == "1"

    def test_stop_times(self) -> None:
        g = self.get_gtfs()
        assert len(g.stop_times) == 304
        assert "100" in g.stop_times

        t = g.stop_times["100"]
        assert len(t) == 19

        st = t[10]
        assert st["trip_id"] == "100"
        assert st["stop_sequence"] == "10"
        assert st["stop_id"] == "micha"
        assert st["arrival_time"] == "19:47:00"
        assert st["departure_time"] == "19:47:00"

    def test_stop_times_by_stop(self) -> None:
        g = self.get_gtfs()
        assert len(g.stop_times_by_stops) == 28
        assert "wsrod" in g.stop_times_by_stops

        s = g.stop_times_by_stops["wsrod"]
        assert len(s) == 296

        st = s[5]
        assert st["trip_id"] == "6"
        assert st["stop_sequence"] == "0"
        assert st["stop_id"] == "wsrod"
        assert st["arrival_time"] == "06:40:00"
        assert st["departure_time"] == "06:40:00"

    def test_stops(self) -> None:
        g = self.get_gtfs()
        assert len(g.stops) == 28
        assert "prusz" in g.stops

        s = g.stops["prusz"]
        assert s["stop_id"] == "prusz"
        assert s["stop_name"] == "Pruszków WKD"
        assert s["stop_lat"] == "52.1616392"
        assert s["stop_lon"] == "20.8166108"
        assert s["wheelchair_boarding"] == "1"

    def test_shapes(self) -> None:
        g = self.get_gtfs()
        assert len(g.shapes) == 10
        assert "5" in g.shapes

        s = g.shapes["5"]
        assert len(s) == 62

        assert s[10] == (52.1046287632, 20.64141690463)

    def test_header_of(self) -> None:
        g = self.get_gtfs()
        assert g.header_of("routes") == [
            "agency_id",
            "route_id",
            "route_short_name",
            "route_long_name",
            "route_type",
            "route_color",
            "route_text_color",
        ]
        assert g.header_of("stop_times") == [
            "trip_id",
            "stop_sequence",
            "stop_id",
            "arrival_time",
            "departure_time",
        ]

    def test_all_dates_of(self) -> None:
        dates = self.get_gtfs().all_dates_of("C")
        assert date(2021, 12, 12) in dates
        assert date(2021, 12, 13) not in dates
        assert date(2021, 12, 24) in dates
        assert date(2022, 12, 30) not in dates
        assert date(2022, 12, 31) in dates


class TestWkdGtfsZip(BaseWkdGtfsTest):
    gtfs_instance: ClassVar[Optional[Gtfs]] = None

    def get_gtfs(self) -> Gtfs:
        if not TestWkdGtfsZip.gtfs_instance:
            TestWkdGtfsZip.gtfs_instance = Gtfs.from_zip(FIXTURE_PATH / "gtfs_wkd.zip")
        return TestWkdGtfsZip.gtfs_instance


class TestWkdGtfsDirectory(BaseWkdGtfsTest):
    gtfs_instance: ClassVar[Optional[Gtfs]] = None

    def get_gtfs(self) -> Gtfs:
        if not TestWkdGtfsDirectory.gtfs_instance:
            TestWkdGtfsDirectory.gtfs_instance = Gtfs.from_directory(FIXTURE_PATH / "gtfs_wkd")
        return TestWkdGtfsDirectory.gtfs_instance


def test_unordered_stop_times() -> None:
    stop_times_file = StringIO(
        (
            "trip_id,stop_sequence,stop_id,arrival_time,departure_time\r\n"
            "t1,1,s0,10:00:00,10:00:00\r\n"
            "t1,2,s1,10:05:00,10:05:00\r\n"
            "t1,3,s2,10:10:00,10:10:00\r\n"
            "t2,3,s0,10:25:00,10:25:00\r\n"
            "t2,2,s1,10:20:00,10:20:00\r\n"
            "t2,1,s2,10:15:00,10:15:00\r\n"
        )
    )

    gtfs = Gtfs()
    gtfs.load_stop_times("stop_times", stop_times_file)

    assert len(gtfs.stop_times) == 2

    assert len(gtfs.stop_times["t1"]) == 3
    assert gtfs.stop_times["t1"][0]["stop_sequence"] == "1"
    assert gtfs.stop_times["t1"][0]["stop_id"] == "s0"
    assert gtfs.stop_times["t1"][1]["stop_sequence"] == "2"
    assert gtfs.stop_times["t1"][1]["stop_id"] == "s1"
    assert gtfs.stop_times["t1"][2]["stop_sequence"] == "3"
    assert gtfs.stop_times["t1"][2]["stop_id"] == "s2"

    assert len(gtfs.stop_times["t2"]) == 3
    assert gtfs.stop_times["t2"][0]["stop_sequence"] == "1"
    assert gtfs.stop_times["t2"][0]["stop_id"] == "s2"
    assert gtfs.stop_times["t2"][1]["stop_sequence"] == "2"
    assert gtfs.stop_times["t2"][1]["stop_id"] == "s1"
    assert gtfs.stop_times["t2"][2]["stop_sequence"] == "3"
    assert gtfs.stop_times["t2"][2]["stop_id"] == "s0"


def test_unordered_shapes() -> None:
    shapes_file = StringIO(
        (
            "shape_id,shape_pt_sequence,shape_pt_lat,shape_pt_lon\r\n"
            "A,2,2.0,2.0\r\n"
            "A,0,0.0,0.0\r\n"
            "A,1,1.0,1.0\r\n"
            "B,3,-3.0,-3.0\r\n"
            "B,1,-1.0,-1.0\r\n"
            "B,2,-2.0,-2.0\r\n"
            "B,0,0.0,0.0\r\n"
        )
    )

    gtfs = Gtfs()
    gtfs.load_shapes("shapes", shapes_file)

    assert len(gtfs.shapes) == 2

    assert len(gtfs.shapes["A"]) == 3
    assert gtfs.shapes["A"][0] == (0.0, 0.0)
    assert gtfs.shapes["A"][1] == (1.0, 1.0)
    assert gtfs.shapes["A"][2] == (2.0, 2.0)

    assert len(gtfs.shapes["B"]) == 4
    assert gtfs.shapes["B"][0] == (0.0, 0.0)
    assert gtfs.shapes["B"][1] == (-1.0, -1.0)
    assert gtfs.shapes["B"][2] == (-2.0, -2.0)
    assert gtfs.shapes["B"][3] == (-3.0, -3.0)



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022-2023 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import csv
import logging
import zipfile
from dataclasses import dataclass, field
from datetime import date, timedelta
from io import TextIOWrapper
from math import nan
from operator import itemgetter
from pathlib import Path
from typing import IO, Callable, List, Optional, Union

from .util import parse_gtfs_date, sequence_to_int

logger = logging.getLogger("jvig.gtfs")

Row = dict[str, str]
Point = tuple[float, float]

TableToOne = dict[str, Row]
TableToMany = dict[str, list[Row]]
TableToPoints = dict[str, list[Point]]
Table = Union[TableToOne, TableToMany, TableToPoints]


def _get_shape_pt(row: Row) -> Optional[Point]:
    """Tries to parse a shapes.txt row into a Point tuple.
    If there's anything wrong with the row, returns None"""
    try:
        return float(row.get("shape_pt_lat", nan)), float(row.get("shape_pt_lon", nan))
    except ValueError:
        return None


_table_keys: dict[str, str] = {
    "agency": "agency_id",
    "stops": "stop_id",
    "routes": "route_id",
    "trips": "trip_id",
    "calendar": "service_id",
    "calendar_dates": "service_id",
    "frequencies": "trip_id",
    "stop_times": "trip_id",
    "shapes": "shape_id",
}


@dataclass
class Gtfs:
    """Gtfs is a class that holds all known GTFS tables.
    It's also responsible for loading the data."""

    agency: TableToOne = field(default_factory=dict)
    stops: TableToOne = field(default_factory=dict)
    stop_children: dict[str, list[str]] = field(default_factory=dict)
    routes: TableToOne = field(default_factory=dict)
    trips: TableToOne = field(default_factory=dict)
    calendar: TableToOne = field(default_factory=dict)
    calendar_dates: TableToMany = field(default_factory=dict)
    frequencies: TableToMany = field(default_factory=dict)
    stop_times: TableToMany = field(default_factory=dict)
    stop_times_by_stops: TableToMany = field(default_factory=dict)
    shapes: TableToPoints = field(default_factory=dict)

    def load_to_row(self, table_name: str, stream: IO[str]) -> None:
        """Loads a table where the key should map into a single row,
        like agency.txt or calendar.txt."""
        primary_key = _table_keys[table_name]
        table: TableToOne = getattr(self, table_name)
        table.clear()

        for row in csv.DictReader(stream):
            # Fix for GTFS feeds without an explicit agency_id
            if table_name in ("agency", "routes") and "agency_id" not in row:
                row["agency_id"] = "(missing)"

            table[row[primary_key]] = row

    def load_to_rows(self, table_name: str, stream: IO[str]) -> None:
        """Loads a table where the key should map into multiple row, like frequencies.txt."""
        primary_key = _table_keys[table_name]
        table: TableToMany = getattr(self, table_name)
        table.clear()

        for row in csv.DictReader(stream):
            table.setdefault(row[primary_key], []).append(row)

    def load_stops(self, table_name: str, stream: IO[str]) -> None:
        """Specialized loader for stops.txt, which loads data into both
        self.stops and self.stop_children."""
        assert table_name == "stops"
        self.stops.clear()
        self.stop_children.clear()

        for row in csv.DictReader(stream):
            self.stops[row["stop_id"]] = row

            # Check if this is a child stop belonging to a larger structure
            parent = row.get("parent_station")
            if parent and row.get("location_type") != "1":
                self.stop_children.setdefault(parent, []).append(row["stop_id"])

    def load_shapes(self, table_name: str, stream: IO[str]) -> None:
        """Specialized loader for shapes.txt to parse the shape."""
        assert table_name == "shapes"

        shapes_with_indices: dict[str, list[tuple[int, Point]]] = {}

        for row in csv.DictReader(stream):
            idx = sequence_to_int(row.get("shape_pt_sequence", ""))
            pt = _get_shape_pt(row)

            # NOTE: Invalid rows are silently ignored
            if idx >= 0 and pt is not None:
                shapes_with_indices.setdefault(row["shape_id"], []).append((idx, pt))

        # Sort shapes by shape_pt_sequence
        for shape in shapes_with_indices.values():
            shape.sort(key=itemgetter(0))

        # Set shapes table
        self.shapes = {
            shape_id: [i[1] for i in shape] for shape_id, shape in shapes_with_indices.items()
        }

    def load_stop_times(self, table_name: str, stream: IO[str]) -> None:
        """Specialized loader for stop_times.txt, which loads the data
        into self.stop_times and self.stop_times_by_stops."""

        assert table_name == "stop_times"
        self.stop_times.clear()
        self.stop_times_by_stops.clear()

        for row in csv.DictReader(stream):
            self.stop_times.setdefault(row["trip_id"], []).append(row)
            self.stop_times_by_stops.setdefault(row["stop_id"], []).append(row)

        # Sort stop_times by stop_sequence
        for trip_stop_times in self.stop_times.values():
            trip_stop_times.sort(key=lambda row: sequence_to_int(row.get("stop_sequence", "")))

    def header_of(self, table_name: str) -> List[str]:
        """Returns the GTFS header of a particlar table."""
        table: Union[TableToOne, TableToMany] = getattr(self, table_name)

        # Get the first entry from the table
        entry = next(iter(table.values()), Row())

        # If it's a list of rows, get its first row
        if isinstance(entry, list):
            entry = entry[0]

        # Return the keys of the row
        return list(entry.keys())

    @property
    def _loader_table(self) -> dict[str, Callable[[str, IO[str]], None]]:
        return {
            "agency": self.load_to_row,
            "stops": self.load_stops,
            "routes": self.load_to_row,
            "trips": self.load_to_row,
            "calendar": self.load_to_row,
            "calendar_dates": self.load_to_rows,
            "frequencies": self.load_to_rows,
            "stop_times": self.load_stop_times,
            "shapes": self.load_shapes,
        }

    @classmethod
    def from_directory(cls, where: Path) -> "Gtfs":
        """Loads GTFS data from a directory of .txt files"""
        self = cls()
        loaders = self._loader_table

        for f in where.glob("*.txt"):
            table_name = f.stem
            loader = loaders.get(table_name)

            if loader:
                logger.info(f"Loading table {table_name}")
                with f.open(mode="r", encoding="utf-8-sig", newline="") as stream:
                    loader(table_name, stream)

        return self

    @classmethod
    def from_zip(cls, where: Path) -> "Gtfs":
        """Loads GTFS data from a .zip archive"""
        self = cls()
        loaders = self._loader_table

        with zipfile.ZipFile(where, mode="r") as archive:
            for f in archive.infolist():
                if not f.filename.endswith(".txt"):
                    logger.warn(f"Unrecognized file in zip: {f.filename}")
                    continue

                table_name = f.filename[:-4]
                loader = loaders.get(table_name)

                if loader:
                    logger.info(f"Loading table {table_name}")
                    with archive.open(f, mode="r") as binary_stream:
                        stream = TextIOWrapper(binary_stream, encoding="utf-8-sig", newline="")
                        loader(table_name, stream)

                else:
                    logger.warning(f"Unrecognized file in zip: {f.filename}")

        return self

    @classmethod
    def from_user_input(cls, where: Path) -> "Gtfs":
        """Loads data from a .zip file (if `where` is a file),
        or from a directory with .txt files (if `where` is not a file)"""
        return cls.from_zip(where) if where.is_file() else cls.from_directory(where)

    def all_stops_in_group(self, stop_id: str) -> list[Row]:
        """Returns all stops in the group to which `stop_id` belongs.

        If the stop doesn't exist, returns an empty list.
        If the stop doesn't belong to a group, returns a list with one element.

        In any other cases, the provided stop will always be the
        first element in the returned list.
        """
        if stop_id not in self.stops:
            return []

        stop = self.stops[stop_id]
        stops = [stop]

        if stop.get("location_type") == "1":
            # This stop is the parent station - just append the children
            stops.extend(
                self.stops[child_id]
                for child_id in self.stop_children.get(stop_id, [])
                if child_id in self.stops
            )

        elif stop.get("parent_station") in self.stop_children:
            # This stop is a child in a station - first append the parent,
            # then the other children
            parent_id = stop["parent_station"]

            if parent_id in self.stops:
                stops.append(self.stops[parent_id])

            stops.extend(
                self.stops[child_id]
                for child_id in self.stop_children.get(parent_id, [])
                if child_id in self.stops and child_id != stop_id
            )

        return stops

    def all_dates_of(self, service_id: str) -> set[date]:
        """Returns a set of all date on which a particular calendar is active"""
        dates: set[date] = set()

        # Read the calendar.txt row
        calendar_row = self.calendar.get(service_id)
        if calendar_row:
            active_weekdays = {
                0: calendar_row["monday"] == "1",
                1: calendar_row["tuesday"] == "1",
                2: calendar_row["wednesday"] == "1",
                3: calendar_row["thursday"] == "1",
                4: calendar_row["friday"] == "1",
                5: calendar_row["saturday"] == "1",
                6: calendar_row["sunday"] == "1",
            }
            day = parse_gtfs_date(calendar_row["start_date"])
            end = parse_gtfs_date(calendar_row["end_date"])
            while day <= end:
                if active_weekdays[day.weekday()]:
                    dates.add(day)
                day += timedelta(days=1)

        # Add the calendar_dates.txt rows
        for dates_row in self.calendar_dates.get(service_id, []):
            day = parse_gtfs_date(dates_row["date"])
            if dates_row["exception_type"] == "1":
                dates.add(day)
            elif dates_row["exception_type"] == "2":
                dates.discard(day)

        return dates



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from markupsafe import escape

from .. import valid

VALID_FIELDS: set[str] = {
    "route_id",
    "agency_id",
    "route_short_name",
    "route_long_name",
    "route_type",
    "route_desc",
    "route_url",
    "route_color",
    "route_text_color",
    "route_sort_order",
}

ROUTE_TYPE_DATA: dict[str, tuple[str, bool]] = {
    "0": ("🚊", False),
    "1": ("🚇", False),
    "2": ("🚆", False),
    "3": ("🚌", False),
    "4": ("⛴️", False),
    "5": ("🚋", False),
    "6": ("🚠", False),
    "7": ("🚞", False),
    "11": ("🚎", False),
    "12": ("🚝", False),
    "100": ("🚆", True),
    "101": ("🚅", True),
    "102": ("🚆", True),
    "103": ("🚆", True),
    "105": ("🚆", True),
    "106": ("🚆", True),
    "107": ("🚂", True),
    "108": ("🚆", True),
    "109": ("🚆", True),
    "200": ("🚌", True),
    "201": ("🚌", True),
    "202": ("🚌", True),
    "204": ("🚌", True),
    "400": ("🚇", True),
    "401": ("🚇", True),
    "402": ("🚇", True),
    "403": ("🚇", True),
    "405": ("🚝", True),
    "700": ("🚍", True),
    "701": ("🚍", True),
    "702": ("🚍", True),
    "704": ("🚍", True),
    "715": ("🚍", True),
    "717": ("🚐", True),
    "800": ("🚎", True),
    "900": ("🚊", True),
    "907": ("🚋", True),
    "1000": ("⛴️", True),
    "1300": ("🚠", True),
    "1400": ("🚞", True),
    "1700": ("❓", True),
}


def header_class(field: str) -> str:
    return "" if field in VALID_FIELDS else "value-unrecognized"


def format_cell(row: dict[str, str], field: str) -> str:
    value = row[field]

    if field in {"route_color", "route_text_color"}:
        if valid.color(value):
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "route_type":
        if value not in ROUTE_TYPE_DATA:
            return f'<td class="value-invalid">{escape(value)}</td>'

        icon, is_extended = ROUTE_TYPE_DATA[value]
        if is_extended:
            return f'<td class="value-extended">{escape(value)} ({icon})</td>'
        else:
            return f"<td>{escape(value)} ({icon})</td>"

    elif field == "route_short_name":
        color = row.get("route_color")
        text_color = row.get("route_text_color")

        # To create a nice color blob we need valid colors
        if not color or not text_color or not valid.color(color) or not valid.color(text_color):
            return f"<td>{escape(value)}</td>"

        style = (
            f"background-color: #{color}; color: #{text_color}; "
            "border-radius: 4px; padding: 2px; margin: 2px;"
        )

        if value == "":
            style += " display: block; width: 14px; height: 14px;"

        return (
            f'<td class="short-name-with-blob"><span style="{style}">'
            f"{escape(value)}</style></td>"
        )

    else:
        return f"<td>{escape(value)}</td>"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import date

import pytest
from jinja2 import Undefined

from jvig import util


def test_time_to_int():
    assert util.time_to_int("10:40:30") == 38430
    assert util.time_to_int("8:00:00") == 28800
    assert util.time_to_int("08:00:00") == 28800
    assert util.time_to_int("26:30:00") == 95400
    assert util.time_to_int("8:20") == -1
    assert util.time_to_int("8") == -1
    assert util.time_to_int("") == -1
    assert util.time_to_int("foo") == -1
    assert util.time_to_int("foo:bar:baz") == -1


def test_sequence_to_int():
    assert util.sequence_to_int("1") == 1
    assert util.sequence_to_int("0") == 0
    assert util.sequence_to_int("454565649") == 454565649
    assert util.sequence_to_int("-5") < 0
    assert util.sequence_to_int("") < 0
    assert util.sequence_to_int("foo") < 0


def test_to_js_literal():
    assert util.to_js_literal(None) == "null"
    assert util.to_js_literal(Undefined()) == "undefined"

    assert util.to_js_literal("") == "''"
    assert util.to_js_literal("foo") == "'foo'"
    assert util.to_js_literal('"foo"') == "'\"foo\"'"
    assert util.to_js_literal('"foo";alert(1)') == "'\"foo\";alert(1)'"

    assert util.to_js_literal(0) == "0"
    assert util.to_js_literal(-2) == "-2"
    assert util.to_js_literal(14564532132) == "14564532132"
    assert util.to_js_literal(-9999999999) == "-9999999999"


def test_parse_gtfs_date():
    assert util.parse_gtfs_date("20200229") == date(2020, 2, 29)
    assert util.parse_gtfs_date("11111111") == date(1111, 11, 11)

    with pytest.raises(ValueError):
        util.parse_gtfs_date("20220229")

    with pytest.raises(ValueError):
        util.parse_gtfs_date("")

    with pytest.raises(ValueError):
        util.parse_gtfs_date("2020")


def test_unique_list():
    assert util.unique_list([1, 2, 2, 3, 3, 1, 4, 5, 5, 4]) == [1, 2, 3, 4, 5]



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import date
from typing import Any, Hashable, Iterable, TypeVar

from jinja2 import is_undefined

T = TypeVar("T")
THashable = TypeVar("THashable", bound=Hashable)


def time_to_int(time: str) -> int:
    try:
        h, m, s = map(int, time.split(":"))
        return h * 3600 + m * 60 + s
    except ValueError:
        return -1


def sequence_to_int(text: str) -> int:
    try:
        return int(text)
    except ValueError:
        return -1


def to_js_literal(obj: Any) -> str:
    if obj is None:
        return "null"

    elif is_undefined(obj):
        return "undefined"

    elif isinstance(obj, str):
        # FIXME: Figure out if this is actually a safe way to pass a string to JS
        return repr(obj)

    elif isinstance(obj, int):
        return str(obj)

    else:
        raise ValueError(f"Unsupported conversion to JS literal from {type(obj).__name__}")


def parse_gtfs_date(x: str) -> date:
    """Parses a GTFS date string ("%Y%m%d") into a datetime.date instance.
    Faster than `datetime.strptime(x, "%Y%m%d").date()`.
    """
    return date(int(x[0:4]), int(x[4:6]), int(x[6:8]))


def unique_list(it: Iterable[THashable]) -> list[THashable]:
    """Similar to `list(it)`, except that duplicate entries are removed.
    In other words, similar to `list(set(it))`, except that order of elements is preserved.
    """
    seen: set[THashable] = set()
    lst: list[THashable] = []

    for i in it:
        if i not in seen:
            seen.add(i)
            lst.append(i)

    return lst



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import quote_plus

from markupsafe import escape

from .. import valid

VALID_FIELDS: set[str] = {
    "trip_id",
    "arrival_time",
    "departure_time",
    "stop_id",
    "stop_sequence",
    "stop_headsign",
    "pickup_type",
    "drop_off_type",
    "shape_dist_traveled",
    "timepoint",
}


def header_class(field: str) -> str:
    return "" if field in VALID_FIELDS else "value-unrecognized"


def format_cell(row: dict[str, str], field: str) -> str:
    value = row[field]

    if field == "trip_id":
        return f'<td><a href="/trip/{quote_plus(value)}">{escape(value)}</a></td>'

    elif field in {"arrival_time", "departure_time"}:
        if valid.time(value):
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "stop_id":
        return f'<td><a href="/stop/{quote_plus(value)}">{escape(value)}</a></td>'

    elif field == "stop_sequence":
        if valid.uint(value):
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field in {"pickup_type", "drop_off_type"}:
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (🚏)</td>"
        elif value == "1":
            return "<td>1 (🚫)</td>"
        elif value == "2":
            return "<td>2 (☎️)</td>"
        elif value == "3":
            return "<td>3 (👈)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "shape_dist_traveled":
        if valid.non_negative_float(value):
            return f"<td>{escape(value)}</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    elif field == "timepoint":
        if value == "":
            return "<td></td>"
        elif value == "0":
            return "<td>0 (🤷)</td>"
        elif value == "1":
            return "<td>1 (📌)</td>"
        else:
            return f'<td class="value-invalid">{escape(value)}</td>'

    else:
        return f"<td>{escape(value)}</td>"



# jvig - GTFS Viewer, created using Flask.
# Copyright © 2022 Mikołaj Kuranowski

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from jvig import valid


def test_color():
    assert valid.color("FFFFFF")
    assert valid.color("bb0000")
    assert valid.color("")
    assert not valid.color("red")
    assert not valid.color("000")


def test_latitude():
    assert valid.latitude("52.14") == 52.14
    assert valid.latitude("-60") == -60.0
    assert valid.latitude("120.15") is None
    assert valid.latitude("") is None
    assert valid.latitude("foo") is None


def test_longitude():
    assert valid.longitude("52.14") == 52.14
    assert valid.longitude("-60") == -60.0
    assert valid.longitude("120.15") == 120.15
    assert valid.longitude("240.8") is None
    assert valid.longitude("") is None
    assert valid.longitude("foo") is None


def test_time():
    assert valid.time("8:20:40")
    assert valid.time("08:20:40")
    assert valid.time("26:00:00")
    assert not valid.time("8:20")
    assert not valid.time("8:0:0")
    assert not valid.time("foo")
    assert not valid.time("")


def test_non_negative_float():
    assert valid.non_negative_float("0")
    assert valid.non_negative_float("1")
    assert valid.non_negative_float("69.420")
    assert valid.non_negative_float("")
    assert not valid.non_negative_float("-69.420")
    assert not valid.non_negative_float("-1")
    assert not valid.non_negative_float("foo")


def test_uint():
    assert valid.uint("0")
    assert valid.uint("1")
    assert not valid.uint("69.420")
    assert not valid.uint("")
    assert not valid.uint("-69.420")
    assert not valid.uint("-0")
    assert not valid.uint("-1")
    assert not valid.uint("foo")


def test_date():
    assert valid.date("20200101")
    assert valid.date("20200229")
    assert not valid.date("20210229")
    assert not valid.date("")
    assert not valid.date("0")
    assert not valid.date("20202020")
    assert not valid.date("20201040")
